/*
 * Copyright (c) 1997 IBM Corporation.
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 * Copyright (c) 1997 ORA, Inc. All Rights Reserved.
 * Code derived from  the original OpenCard Framework
 */
package com.ora.smartcard.terminal.reflex20;

import java.util.Properties;
import java.util.Vector;

import opencard.terminal.CardTerminal;
import opencard.terminal.CardTerminalException;
import opencard.terminal.CardTerminalIOControl;
import opencard.terminal.CardTerminalIOBlender;
import opencard.terminal.Slot;
import opencard.terminal.CardRequest;
import opencard.terminal.UserInteraction;
import opencard.terminal.UserInteractionHandler;
import opencard.agent.ResponseAPDU;
import opencard.agent.CommandAPDU;
import opencard.agent.CardAgentException;
import opencard.io.SmartCard;

/** Implementation of an OpenCard <tt>CardTerminal</tt> for Reflex20.
  *
  * @see opencard.terminal.CardTerminal
  */
public class Reflex20CardTerminal extends CardTerminal implements Reflex20Constants {
  private static boolean TRACE = Boolean.getBoolean("com.sun.smartcard.terminal.reflex20.trace");
  private Reflex20Driver driver = null;

  /** Instantiate an <tt>Reflex20Terminal</tt>.
    *
    * @param  name
    *         The user friendly name.
    * @param  type
    *         The terminal type (here "Reflex20")
    * @param  address
    *         A <tt>String</tt> naming the port the terminal is connected to.
    */
  protected Reflex20CardTerminal(String name, String type, 
				String address) 
    throws CardTerminalException {
      super(name, type, address);

      slotArray = new Slot[1];

      configConnector(address);
      slotArray[0] = new Reflex20Slot(this, 0, driver);
  }

  /** Display a message.<p>
    *
    * @param   message
    *          The <tt>String</tt> to display.
    */
  public void display(String message) {
    int length = (message.length() > 127) ? 127 : message.length();
    byte[] command = new byte[length + 1];

    command[0] = (byte)length;
    byte[] text = message.getBytes(); // define a character encoding here
    System.arraycopy(text, 0, command, 1, text.length);

    try {
      controlCommand(new CommandAPDU(TERMINAL, DISPLAY, (byte)0, (byte)0,
				     command));
    }
    catch (CardTerminalException e) {
      // just ignore it???
    }

    return;
  }

  /** Clear the display.<p>
    */
  public void clearDisplay() {
    byte[] command = {1, 0xc};

    try {
      controlCommand(new CommandAPDU(TERMINAL, DISPLAY, (byte)0, (byte)0,
				     command));
    }
    catch (CardTerminalException e) {
      // just ignore it ???
    }

    return;
  }

  /** Return keyboard (i.e., keyboard, PIN pad, etc.) input.<p>
    * 
    * @param  ioControl
    *         A <tt>CardTerminalIOControl</tt> object that specifies
    *         how the user input should look like.
    * @return A <tt>String</tt> containing the input.
    */
  public String keyboardInput(CardTerminalIOControl ioControl) {
		throw new CardTerminalException("No Key Board");
  }

  /** Read one key from the pin pad.
    * The mapping to symbolic names is defined in <tt>Reflex20Constants</tt>.
    * @see com.ibm.zurich.smartcard.terminal.Reflex20Constants
    *
    * @param     maxChars
    *            Number of characters to read.
    * @param     timeout
    *            Time to wait between two key strokes
    * @return    the character read.
    * @exception opencard.terminal.CardTerminalException 
    *            Thrown when timeout for a key stroke was exceeded
    */
  private String readPinPad(int maxChars, int timeout) throws CardTerminalException {
      throw new CardTerminalException("No Pin Pad");
  }

  private char keyToChar(byte key) throws CardTerminalException { 
      throw new CardTerminalException("Unexpected Character");
  }

  private String[] reflex20Features = {
    "display", "false",
    "display.type", "String",
    "display.lines", "0",
    "display.columns", "0",
    "keyboard", "false",
    "keyboard.type", "pinpad",
    "keyboard.keys", "0,1,2,3,4,5,6,7,8,9,0,C,OK,F1,F2,F3,F4",
    "leds", "0",
    "led.0", "orange",
    "led.1", "green",
    "removable", "yes",
  };

  /** Query the card terminal about its features.<p>
   *
   * @return A list of properties describing the features of this
   *         card terminal.
   */
  public Properties internalFeatures(Properties features) {
    // ... add Reflex20 specific features
    for (int i = 0; i < reflex20Features.length; i += 2) 
      features.put(reflex20Features[i], reflex20Features[i+1]);
    return features;
  }

  /** Prompt the user for a password, etc.<p>
    * This method displays a message and reads a password from
    * the user. The implementation of this method depends on the
    * ioBlender implementation of the terminal. E.g. a PIN input
    * could be implemented by the terminals numerical keyboard
    * whereas an alphanumeric password might be implemented through
    * as <tt>UserInteractionHandler</tt>.
    *
    * @param  prompt
    *         The message to be displayed.
    * @param  ioControl
    *         A <tt>CardTerminalIOControl</tt> object that specifies
    *         how the user input should look like.
    * @return A <tt>String</tt> containing the input.
    */
  public String promptUser(String prompt, CardTerminalIOControl ioControl) {
		throw new CardTerminalException("Could not reset the terminal");
	}

  /** Issue a card terminal specific command.<p>
   *
   * This method is almost as generic as a method can get due to the nature of
   * a card terminal <i>specific</i> command. Similar to the way the Java
   * SmartCard API supports application specific commands through ASC classes
   * and a the associated lookup methods, one could provide card specific
   * classes (CSC classes that offer a specific API) together with a lookup
   * method in this class.<p>
   *
   * @param     appSpecCmd
   *            A string containing the card terminal specific command.
   * @param     appSpecParameter
   *            An object encapsulating any required parameters for the 
   *            card terminal specific command.
   * @param     timeout
   *            The time to wait before returning (in seconds); <tt>-1</tt> signals 
   *            an indefinite timeout (i.e., wait forever).
   * @return    An object that encapsulates the return result of the card
   *            terminal specific command.  
   * @exception opencard.terminal.CardTerminalException
   *            Thrown when a card is not present.
   */
  public Object command(String appSpecCmd, Object appSpecParameter, int timeout) 
    throws CardTerminalException {
      return null;
  }

  protected void configConnector(String connector) throws CardTerminalException {
    int port = Integer.parseInt(connector);

    driver = new Reflex20Driver(port, port); //logical number equals port
    driver.init();
    reset();
  }

  /** Reset the terminal.
   *
   * @exception opencard.terminal.CardTerminalException 
   *            Did not reset properly.
   */
  protected void reset() throws CardTerminalException {

	// Send a reset without asking for an ATR
    ResponseAPDU resp = controlCommand(new CommandAPDU(RESETCLASS, RESET,
						       (byte)0, (byte)0,
						       new byte[] { 0 }));

    switch (resp.sw()) {
      case 0x9000:
      case 0x9001:
      case 0x64A1:
      case 0x64A2:
        return;
      case 0x62A5:
      case 0x62A6:
      case 0x62A7:
      case 0x64A3:
      case 0x64A8:
        throw new CardTerminalException("APDU error:" + resp, this);
      default:
        throw new CardTerminalException("Unknown response: "+ resp.sw(), this, resp.sw());
    }
  }


  /** Send control command to terminal. Uses APDU classes.
    *
    * @param     command 
    *            The APDU to be sent to the terminal.
    * @return    Response from terminal.
    * @exception opencard.CardTerminalException 
    *            Exception thrown by driver.
    */
  protected ResponseAPDU controlCommand(CommandAPDU command)
    throws CardTerminalException {
      return new ResponseAPDU(this.driver.data((byte)1,
					       command.toT1ByteArray()));
  }

  /** Wait for card insertion and return a <tt>SmartCard</tt> object.<p>
    *
    * @param    cr
    *           The <tt>CardRequest</tt> describing the characteristics
    *           of this request.
    * @return    A <tt>SmartCard</tt> object representing the inserted smart card.
    * @exception opencard.terminal.CardTerminalException
    *            Signals a time out condition.
    */
  public SmartCard waitForCard(CardRequest cr) throws CardTerminalException {
    return slotArray[0].waitForCard(cr);
  }

  /** Release driver resources when garbage collected.
    */
  protected void finalize() {
    try {
      driver.close();
    }
    catch (CardTerminalException e) {
      // ignore
    }
  }
}

