/*
 * Copyright (c) 1997 IBM Corporation.
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 * Copyright (c) 1997 ORA, Inc. All Rights Reserved.
 * Code derived from  the original OpenCard Framework
 */
package com.ora.smartcard.terminal.reflex20;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import opencard.terminal.CardTerminalException;

/** Driver for Reflex20 Card Terminal which is a PCCard.
 * Provide access to native methods of CT_* functions.
 */

public final class Reflex20Driver implements Reflex20Constants {
  /** TRACE flag.
    */
  private static boolean TRACE = Boolean.getBoolean("com.sun.ora.smartcard.terminal.reflex20.trace");
  /** Logical terminal number.
    */
  private int ctNumber;

  /** Communication port number.
    */
  private int portNumber;

  /**-- Misc. privates
    */
  private final int responseMax = 512;
  private byte[] response;
  private int responseLength;

  /** The constructor.<p>
    * 
    * @param ctNumber Logical terminal number.
    * @param portNumber Number of communication port for terminal.
    */
  Reflex20Driver(int ctNumber, int portNumber) {
    this.ctNumber = ctNumber;
    this.portNumber = portNumber;
    response = new byte[responseMax];
  }
  
  /** Initialize the protocol layer on the workstation,
    * the Reflex20 card terminal and the communication port 
	* (see Reflex20 documentation).
    *
    * @exception com.com.ibm.smartcard.terminal.CardTerminalException 
    *            Error code from native code.
    */
  void init() throws CardTerminalException {
    int rc = ctInit(ctNumber, portNumber);
    
    if (rc == CT_OK)
      return;
    else
      throw new CardTerminalException(rc2String(rc));
  }

  /** Send commands and data to the card terminal or Smart Card and
    * receive response.<p>
    *
    * @param destination Send to card terminal or ICC.
    * @param pdu Data to send.
    * @return Response APDU.
    * @exception com.com.ibm.smartcard.terminal.CardTerminalTException Error code from native code.
    */
  synchronized byte[] data(byte destination, byte[] pdu) throws CardTerminalException {
    int rc = ctData(destination, pdu, pdu.length, response, response.length);
    
    if (rc == CT_OK) {
      byte[] result = new byte[responseLength];
      System.arraycopy(response, 0, result, 0, responseLength);
      return result;
    }
    else
      throw new CardTerminalException(rc2String(rc));
  }

  /** Terminate communication with terminal.<p>
    *
    * @exception com.com.ibm.smartcard.terminal.CardTerminalException Error code from native code.
    */
  public void close() throws CardTerminalException {
    int rc = ctClose(ctNumber);

    if (rc == CT_OK)
      return;
    else
      throw new CardTerminalException(rc2String(rc));
  }

  /** Human readable error message.
    *
    * @param    rc 
    *           error code
    * @return   description
    */
  protected String rc2String(int rc) {
    switch(rc) {
    case CT_ERR_INVALID:
      return "invalid parameter (CT_ERR_INVALID)";
    case CT_ERR_CT:
      return "communication error (CT_ERR_CT)";
    case CT_ERR_TRANS:
      return "uncorrectable transmission error (CT_ERR_TRANS)";
    case CT_ERR_MEMORY:
      return "out of memory (CT_ERR_MEMORY)";
    case CT_ERR_HTSI:
      return "internal error (CT_ERR_HTSI)";
    default:
      return "unexpected return code "+rc;
    }
  }

  // The native methods to access the CT_* API
  private static synchronized native int ctInit(int ctNumber, int portNumber);
  private native int ctData(byte destination, byte[] pdu, int pduLength,
			    byte[] response, int responseLength);
  private static synchronized native int ctClose(int ctNumber);

  static {
    String vendor = System.getProperty("java.vendor");

    if (vendor != null && vendor.equals("Netscape Communications Corporation")) {
      // Actually, we should check which version of Communicator we are running on.
      // The java.version property does not offer much help here. It was the same
      // for both Navigator 3.01 and early versions of Communicator.
      try {

	// This is what we wanted to do... but it requires netscape classes
	// 	PrivilegeManager.enablePrivilege("UniversalLinkAccess");
	//       } catch (ForbiddenTargetException e) {
	// 	System.err.println("Access denied: " + e);
	// 	e.printStackTrace();
	//       }

	// So go for the less efficient way and use reflection...
 	Class[] parameterTypes = { java.lang.String.class };
	
 	Class privilegeManager = Class.forName("netscape.security.PrivilegeManager");
 	Method enablePrivilege = privilegeManager.getMethod("enablePrivilege",
 							    parameterTypes);
 	Object[] parameter = { "UniversalLinkAccess" };
 	enablePrivilege.invoke(null, parameter);
      } catch (ClassNotFoundException e) {
   	System.err.println("Reflex20Driver: PrivilegeManager class not found");
	if (TRACE)
	  e.printStackTrace();
      } catch (NoSuchMethodException e) {
   	System.err.println("Reflex20Driver: method enablePrivilege not found");
	if (TRACE)
	  e.printStackTrace();
      } catch (IllegalAccessException e) {
   	System.err.println("Reflex20Driver: enablePrivilege illegal access");
	if (TRACE)
	  e.printStackTrace();
      } catch (InvocationTargetException e) {
	if (TRACE)
	  System.err.println("Reflex20Driver: enablePrivilege failed");
   	e.getTargetException().printStackTrace();
      }
    }
      //      System.loadLibrary("Reflex20DriverJRI");
      //    } else {
    System.loadLibrary("Reflex20Driver_w32");
  }
}
