/*
 * Copyright (c) 1997 IBM Corporation.
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 * Copyright (c) 1997 ORA, Inc. All Rights Reserved.
 * Code derived from  the original OpenCard Framework
 */
package com.ora.smartcard.terminal.reflex20;

import opencard.terminal.CardTerminal;
import opencard.terminal.PollingSlot;
import opencard.terminal.CardID;
import opencard.terminal.CardTerminalException;
import opencard.terminal.CardRequest;
import opencard.agent.CardAgentException;
import opencard.io.SmartCard;
import opencard.agent.CommandAPDU;
import opencard.agent.ResponseAPDU;

/** A <tt>Slot</tt> object represents an individual smart card slot of a 
 * card terminal. An application can obtain a <tt>SmartCard</tt> object
 * through the <tt>Slot</tt> object by calling one of the <tt>waitForCard()</tt>
 * methods.
 *
 * @see opencard.terminal.PollingSlot
 * @see opencard.terminal.Slot
 */
public final class Reflex20Slot extends PollingSlot implements Reflex20Constants {
  private static boolean TRACE = Boolean.getBoolean("com.ora.smartcard.terminal.reflex20.trace");
  private static int MAX_TIMEOUT = 60;
  private CardID cid = null;
  private byte slotID;
  private Reflex20Driver driver = null;
  byte[] nullCommand = new byte[] {0};

  /** Instantiate a <tt>Slot</tt> object belonging to the <tt>CardTerminal</tt>
   * object <tt>cardTerminal</tt>.
   *
   * @param   cardTerminal
   *          The card terminal to which this <tt>Slot</tt> belongs to.
   * @param   slotID 
   *          The logical number of the <tt>Slot</tt>.
   * @param   driver 
   *          The <tt>Reflex20Driver</tt> that actually send the APDUs.
   */
  public Reflex20Slot(CardTerminal cardTerminal, int slotID,
		     Reflex20Driver driver) {
    super(cardTerminal);
    this.slotID = (byte)slotID;
    this.driver = driver;
  }

  /** Check whether there is a smart card present.
   * 
   * @return True if there is a smart card inserted in the card 
   *         terminals slot.
   */
  public boolean cardPresent() {
    try {
      byte[] result = getTerminalStatus();

      return (result[0] & 1) != 0;
    } catch (CardTerminalException e) {
      cid = null;
      return false;
    }
  }

  /** Power up an inserted smartcard.<p>
    * If the card was already powered up it is resetted.
    *
    * @param     timeout
    *            The time to wait before returning (in seconds); <tt>-1</tt> signals 
    *            an indefinite timeout (i.e., wait forever).
    * @exception opencard.terminal.CardTerminalException
    *            Thrown when no card is present.
    */
  public void powerUpCard(int timeout) throws CardTerminalException {
    boolean timedOut;
    if (timeout == -1) {	// infinite timeout...
      do {
	    timedOut = requestCard(MAX_TIMEOUT);
      } while (timedOut);
    } else {
      if (timeout > MAX_TIMEOUT) {
	    for ( ; timeout > MAX_TIMEOUT; timeout -= MAX_TIMEOUT)
	      if(!requestCard(MAX_TIMEOUT))
	        return;
      }
      if (requestCard(timeout))
	    throw new CardTerminalException("operation timed out", this);
    }
  }

  /** Power power down an inserted smartcard.<p>
    * If the card was already powered down the method has no effect.
    *
    * @param     timeout
    *            The time to wait before returning (in seconds);
    *            <tt>-1</tt> signals an indefinite timeout (i.e., wait
    *            forever).
    * @exception opencard.terminal.CardTerminalException
    *            Thrown when no card is present.
    */
  public void powerDownCard(int timeout) throws CardTerminalException {
    if (!cardPresent())
      throw new CardTerminalException("no card present", this);

    ResponseAPDU rp = ejectCard(timeout);

    switch(rp.sw()) {
    case 0x9000:
    case 0x9001:
    case 0x6200:
      if (TRACE)
	System.err.println("Card powered down successfully");
      break;
      
    default:
      throw new CardTerminalException("powerDown: unknown status code (" +
				      rp.sw() + ")", this, rp.sw());
    }
  }

  /** Perform a CHV on the card.<p>
    *
    * @param prompt
    *        message to be displayed in the display
    * @param verifyCommand
    *        command APDU that performs the PIN verification
    * @param pinLength
    *        number of characters in PIN
    * @param pinOffset
    *        offset of the PIN within the verifyCommand
    * @param pinType
    *        coding of the PIN, currently "ASCII" or "BCD"
    * @param timeout
    *        timeout for the operation in seconds
    * @param ticket
    *        security ticket
    * @exception opencard.terminal.CardTerminalException 
    *            Card terminal problems.
    */
  private ResponseAPDU verifyCardHolder(String prompt, CommandAPDU verifyCommand,
					int pinLength, int pinOffset,
					String pinType, int timeout,
					Object ticket)
    throws CardTerminalException, IllegalAccessException {
	throw new CardTerminalException("Unimplemented Exception");
  }

  /** Perform a command sequence (send APDU and return card response)
    *
    * @param     data
    *            APDU to be sent to the card.
    * @param     timeout
    *            The time to wait before returning (in seconds); <tt>-1</tt> signals 
    *            an indefinite timeout (i.e., wait forever).
    * @return    Card response.
    * @exception opencard.terminal.CardTerminalException 
    *            Card terminal problems.
    */
  protected ResponseAPDU commandResponse(CommandAPDU data, int timeout) throws CardTerminalException {
    if (TRACE) System.err.println("Zueri Slot sending: " + data);
    byte[] result = driver.data(slotID, data.toT1ByteArray());
    if (TRACE) System.err.println("Zueri Slot done.");
    
    return new ResponseAPDU(result);
  }

  /** Wait for card insertion.<p>
   *
   * @param     cr
   *            The <tt>CardRequest</tt> defining the characteristics of the
   *            request.
   * @return    A <tt>CardID</tt> object representing the inserted smart card.
   */
  protected CardID waitForCardID(int timeout) throws CardTerminalException {

    byte[] result = getTerminalStatus();
    
    if ((result[0] & 3) == 3) { // card present and powered up
      if (cid == null) {
	if (TRACE)
	  System.err.println("Internal error CardID lost");
	throw new CardTerminalException("CardID lost");
      } else {
	return cid;
      }
    }
    powerUpCard(timeout);

    return cid;
  }

  protected CardID waitForCardID()
    throws CardTerminalException {
      return waitForCardID(-1);
  }

  /** Wait for card insertion, power up the card, and perform reset.
    *
    * @param timeout
    *        Timeout in seconds (1..60).
    * @return true if no card has been found within the timeout.
    * @exception opencard.terminal.CardTerminalException 
    *            Card did not reset properly.
    */
  private boolean requestCard(int timeout) throws CardTerminalException {
    CommandAPDU cmdAPDU = null;
    cid = null;			// invalidate old card id
	// Prepare a warm reset command (see Reflex 20 Developer's Handbook
    cmdAPDU = new CommandAPDU(RESETCLASS, WRESET, PATR, PATR, nullCommand ); 
    ResponseAPDU resp = controlCommand( cmdAPDU );
    switch (resp.sw()) {
      case 0x9000:
      case 0x9001:
        cid = new CardID(this, resp.body());
        return false;
	  case 0x64a1:
      case 0x64A2:
        return true;
      case 0x62A5:
      case 0x62A6:
      case 0x62A7:

      case 0x64A3:
      case 0x64A6:
      case 0x64A8:
        throw new CardTerminalException("card reset failed");
      default:
        throw new CardTerminalException("Unknown Response APDU" + resp , this);
      }
   }
  /** "Eject" the card. Actually it is not ejected but we can wait for
    * the card to be removed from the terminal.
    *
    * @param timeout
    *        Time to wait for card removal (in seconds).
    * @return the answer as read from the terminal.
    */
  private ResponseAPDU ejectCard(int timeout) throws CardTerminalException {
    byte[] command = new byte[2];

    if (timeout < 0 || timeout > 60)
      timeout = 60;

    if (timeout == 0) timeout = 1;

    command[0] = 1;
    command[1] = (byte)timeout;

    return controlCommand(new CommandAPDU(TERMINAL, EJECTCARD, (byte)1,
					  (byte)0, command));
  }

  /** Request status of terminal.
    *
    * @return    Status information in some non-standard format.
    * @exception ibm.smartoken.terminal.CardTerminalException
    *            Exceptions thrown by driver.
    */
  private byte[] getTerminalStatus() throws CardTerminalException {
 	CommandAPDU cmdAPDU = null;

	// Prepare a file command (see Reflex 20 Developer's Handbook
    byte[] command = new byte[3];
	command[0] = 0x02;
	command[1] = 0x70;
    command[2] = 0x21;

 	cmdAPDU = new CommandAPDU( ICC, (byte)0xA4, Pone0, Ptwo0, command);

    ResponseAPDU result = controlCommand( cmdAPDU );
    
    switch (result.sw()) {
      case 0x9000:
      case 0x9001:
      case 0x6282:
        break;

      case 0x64A1:
      case 0x6200:
      case 0x6400:
        throw new CardTerminalException("Card Select fle failed");
      default:
        throw new CardTerminalException("Unknown Response APDU" + result , this);
    }
    cmdAPDU = new CommandAPDU( ICC, RBINARY , Pone0, Ptwo0, nullCommand );
    result = controlCommand( cmdAPDU );
    switch (result.sw()) {
       case 0x9000:
       case 0x9001:
     	return result.body();
       case 0x6200:
       case 0x6400:
         throw new CardTerminalException("Read Binary failed");
       default:
        throw new CardTerminalException("Unknown Response APDU" + result , this);
     }
  }

  /** Send control command to terminal.
    *
    * @param command
    *            Command and Le field.
    * @return    Response from terminal.
    * @exception opencard.CardTerminalException 
    *            Exception thrown by driver.
    */
  private ResponseAPDU controlCommand(CommandAPDU command) throws CardTerminalException {
      return new ResponseAPDU(this.driver.data((byte)1, command.toT1ByteArray()));
  }
  /** Parse the reponse in Tag format
    *
    * @param response
    *            Command and Le field.
    * @return    Tag Value as an object
    * @exception opencard.CardTerminalException 
    *            Exception thrown by driver.
    */
}
